import itertools

import numpy as np
from pyDOE2 import lhs


def lhsdesign(n, min_range, max_range, k=5, include_vertices=False):
    """Returns the Latin Hypercube Sampling for a given range of values.

    Parameters
    ----------
    n : int
        Number of samples of the hypercube.
    min_range : np.array
        1-by-p or p-by-1 array containing the minimum values for each variable.
    max_range : np.array
        1-by-p or p-by-1 array containing the maximum values for each variable.
    k : int, optional
        Number of iterations to attempt to improve the design.
    include_vertices : bool
        To include or not the vertices of the hypercube in the sample.

    Returns
    -------
    out : np.array
        n-by-p array containing the Latin Hypercube Sampling.

    Raises
    ------
    ValueError
        If ndim of either `min_range` or `max_range` is not 2.

        If the `min_range` or `max_range` aren't vectors.


    """

    # check input ranges dimensions. If ndim != 1, raise error
    if min_range.ndim != 1 or max_range.ndim != 1:
        raise ValueError("Input ranges must be 1D arrays.")
    else:
        # both have ndim == 1, check if they have the same size
        if min_range.size != max_range.size:
            raise ValueError("min_range and max_range must have the same number of elements")

        # min_range = min_range.reshape(1, -1)
        # max_range = max_range.reshape(1, -1)

        p = min_range.size

    # proceed with normal calculations
    slope = np.tile(max_range - min_range, (n, 1))
    offset = np.tile(min_range, (n, 1))

    # create normalized LH
    x_normalized = lhs(p, samples=n, iterations=k, criterion='maximin')

    if include_vertices:
        vertices = get_vertices(min_range, max_range)

        # scale and return the LH
        return np.vstack((x_normalized * slope + offset, vertices))
    else:
        # scale and return the LH
        return x_normalized * slope + offset


def get_vertices_index(n):
    return np.asarray(list(itertools.product('01', repeat=n))).astype(np.int)


def get_vertices(lb, ub):
    """ Returns all vertices of the hypercube.

    Parameters
    ----------
    lb : np.array
        Lower bound of the hypercube.
    ub : np.array
        Upper bound of the hypercube.

    Returns
    -------
    out : np.array
        2**n-by-n array containing all the vertices of the n-dimensional hypercube. Each line corresponds to a vertex
        of the hypercube

    Raises
    ------
    ValueError
        if the bounds ´lb´ or ´ub´ does not have same shape or number of elements in each

    """
    if lb.ndim == 1 or ub.ndim == 1:
        lb = lb.reshape(1, -1)
        ub = ub.reshape(1, -1)

    if lb.shape != ub.shape:
        raise ValueError(f"´lb´ and ´ub´ must have the same number of elements. ´lb´ = {lb.size} and ´ub´= {ub.size}")

    n = lb.shape[1]  # number of dimensions
    m = 2 ** n  # number of vertices
    vertices_index = get_vertices_index(n)

    cat_placeholder = np.vstack((lb, ub))

    vertices = np.zeros(vertices_index.shape)
    for j in np.arange(n):
        for i in np.arange(m):
            vertices[i, j] = cat_placeholder[vertices_index[i, j], j]

    return vertices


import numpy as np


class BoxMin:
    """BoxMin optimizer used to find the hyperparameters of the Dace class 
    (fit).
    """

    def __init__(self, dace_object, theta0: np.ndarray,
                 theta_lower: np.ndarray, theta_upper: np.ndarray):
        # initialize
        self.dace = dace_object
        self._start(theta0, theta_lower, theta_upper)

        if not np.isinf(self.f):
            # Iterate
            p = self.t.size
            if p <= 2:
                kmax = 2
            else:
                kmax = np.minimum(p, 4)

            for k in np.arange(kmax):
                th = self.t.copy()
                self._explore()
                self._move(th)

        self.perf_info = {'nv': self.nv,
                          'perf': self.perf[:, 0:self.nv]}

    def _start(self, t0: np.ndarray, lo: np.ndarray, up: np.ndarray):
        # get starting point and iteration parameters
        t = t0.reshape(-1, 1)
        lo = lo.reshape(-1, 1)
        up = up.reshape(-1, 1)
        p = t.size

        D = 2 ** (np.arange(1, p + 1).reshape(-1, 1) / (p + 2))

        ee = np.nonzero(np.equal(up, lo))  # Equality constraints
        if ee[0].size != 0:
            D[ee] = np.ones((ee[0].size, 1))
            t[ee] = up[ee]

        # Free starting values
        ng = np.logical_or(np.less(t, lo), np.less(up, t))
        ng = (np.nonzero(ng))

        t[ng] = (lo[ng] * up[ng] ** 7) ** (1 / 8)  # Starting point

        ne = np.nonzero(D != 1)

        # Check starting point and initialize perfomance info
        f = self.dace._objfunc(t)
        nv = 1

        # itpar
        self.D = D
        self.ne = ne
        self.lo = lo
        self.up = up
        self.perf = np.zeros((p + 2, 200 * p))
        self.nv = nv
        self.perf[:, [0]] = np.vstack((t, f, 1))

        # other params
        self.f = f
        self.t = t

        if np.isinf(f):  # bad parameter region
            return

        if ng[0].size > 1:  # Try to improve starting guess
            d0 = 16
            d1 = 2
            q = ng[0].size
            th = t.copy()
            fh = f
            jdom = ng[0][0]

            for k in np.arange(q):
                j = ng[0][k]
                fk = fh
                tk = th

                DD = np.ones((p, 1))
                DD[ng[0]] = np.tile(1 / d1, (q, 1))
                DD[j] = 1 / d0

                alpha = np.min(
                    np.log(lo[ng[0]] / th[ng[0]]) / np.log(DD[ng[0]])) / 5
                v = DD ** alpha
                tk = th

                for rept in np.arange(4):
                    tt = tk * v
                    ff = self.dace._objfunc(tt)
                    nv += 1
                    self.perf[:, [nv - 1]] = np.vstack((tt, ff, 1))

                    if ff <= fk:
                        tk = tt.copy()
                        fk = ff

                        if ff <= f:
                            t = tt.copy()
                            f = ff
                            jdom = j.copy()

                    else:
                        self.perf[[-1], [nv - 1]] = -1
                        break

            # Update data
            if jdom > 0:
                D[np.hstack((0, jdom - 1))] = D[np.hstack((jdom - 1, 0))]
                self.D = D

        self.nv = nv

        # other params
        self.f = f
        self.t = t

    def _explore(self):
        # explore step

        nv = self.nv
        ne = self.ne

        t = self.t
        f = self.f

        for k in np.arange(ne[0].size):
            j = ne[0][k]
            tt = t.copy()
            DD = self.D[j]

            if t[j] == self.up[j]:
                atbd = True
                tt[j] = t[j] / np.sqrt(DD)

            elif t[j] == self.lo[j]:
                atbd = True
                tt[j] = t[j] * np.sqrt(DD)

            else:
                atbd = False
                tt[j] = np.minimum(self.up[j], t[j] * DD)

            ff = self.dace._objfunc(tt)
            nv += 1
            self.perf[:, [nv - 1]] = np.vstack((tt, ff, 2))

            if ff < f:
                t = tt.copy()
                f = ff

            else:
                self.perf[-1, [nv - 1]] = -2

                if not atbd:  # try decrease

                    tt[j] = np.maximum(self.lo[j], t[j] / DD)
                    ff = self.dace._objfunc(tt)
                    nv += 1
                    self.perf[:, [nv - 1]] = np.vstack((tt, ff, 2))

                    if ff < f:
                        t = tt.copy()
                        f = ff

                    else:
                        self.perf[-1, [nv - 1]] = -2

        self.nv = nv

        # other params
        self.f = f
        self.t = t

    def _move(self, th: np.ndarray):
        # pattern move
        nv = self.nv

        t = self.t
        f = self.f

        p = t.size

        v = t / th

        if np.all(v == 1):
            self.D = self.D[np.r_[1:p, 0]] ** 0.2
            return

        # proper move
        rept = True
        while rept:
            tt = np.minimum(self.up, np.maximum(self.lo, t * v))
            ff = self.dace._objfunc(tt)
            nv += 1
            self.perf[:, [nv - 1]] = np.vstack((tt, ff, 3))

            if ff < f:
                t = tt.copy()
                f = ff
                v = v ** 2

            else:
                self.perf[-1, [nv - 1]] = -3
                rept = False

            if np.any(np.logical_or(np.equal(tt, self.lo),
                                    np.equal(tt, self.up))):
                rept = False

        self.nv = nv
        self.D = self.D[np.r_[1:p, 0]] ** 0.25

        # other params
        self.f = f
        self.t = t



import numpy as np
from scipy.linalg import LinAlgError, cholesky, qr
from scipy.spatial.distance import pdist

from .correlation import corr
from .regression import regrpoly
from .utils.matrixdivide import mldivide, mrdivide
from .utils.optimizers import BoxMin


class Dace:
    """This the object oriented implementation of the DACE toolbox MATLAB 
    implementation.

    Parameters
    ----------
    regression : str, optional
        Type of mean regression model. Valid models are:

            * Zero order polynomial ('poly0').
            * First order polynomial ('poly1').
            * Second order polynomial ('poly2').

    correlation : str, optional
        Correlation model. Only Gaussian correlation ('corrgauss') implemented.

    optimizer : str, optional
        Type of NLP optimizer used to find the hyperparameters `theta`. Valid 
        optimizers are  'boxmin'. Default is 'boxmin' (original 
        implementation).
    """

    def __init__(self, regression: str, correlation: str,
                 optimizer: str = 'boxmin'):
        self._S = None
        self._Y = None
        self._regression = regression
        self._correlation = correlation
        self._optimizer = optimizer

        self._sampling_method = 'simplicial'
        self._n_points = 100

    # ------------------------------ PROPERTIES -------------------------------
    @property
    def S(self) -> np.ndarray:
        """m-by-n array of design sites (m being the number of samples and n
        the number of input dimensions."""
        return self._S

    @S.setter
    def S(self, value: np.ndarray):
        # normalize the data
        mS = np.mean(value, axis=0)
        sS = np.std(value, axis=0, ddof=1)

        sS[sS == 0] = 1.0

        self._S = (value - mS) / sS

        self.m, self.n = self._S.shape
        self._Ssc = np.vstack((mS, sS))

    @property
    def Y(self) -> np.ndarray:
        """m-by-q array of observed responses (q is the output dimensions).
        q = 1 mean univariate model, otherise the model is considered
        multivariate."""
        return self._Y

    @Y.setter
    def Y(self, value: np.ndarray):
        if value.shape[0] == 1 or value.ndim == 1:
            # if the Y array is row vector or 1D array make it a column vector
            value = value.reshape(-1, 1)

        m, _ = value.shape

        if m != self.S.shape[0]:
            raise ValueError('S and Y must have the same number of rows')

        # normalize the data
        mY = np.mean(value, axis=0)
        sY = np.std(value, axis=0, ddof=1)

        sY[sY == 0] = 1.0

        self._Y = (value - mY) / sY
        self._Ysc = np.vstack((mY, sY))

    # ---------------------------- PRIVATE METHODS ----------------------------
    def _initialize(self) -> dict:
        """Initialize the parameters for the objective function.
        """
        if self.S is None or self.Y is None:
            raise AttributeError("S and/or Y arrays not defined.")

        S = self.S
        Y = self.Y
        sS = np.std(S, axis=0, ddof=1)
        sS[sS == 0] = 1.0

        m, n = S.shape

        # calculate distances D between points
        mzmax = int(m * (m - 1) / 2)  # number of non-zero distances
        D = np.zeros((mzmax, n))  # initialize matrix with distances

        # create indexes for distances between points
        for k in range(n):
            D[:, k] = pdist(S[:, [k]], metric='euclidean')

        if np.min(np.sum(np.abs(D), axis=1), axis=0) == 0:
            raise ValueError('Multiple design sites are not allowed.')

        # regression matrix
        F = regrpoly(S, polynomial=self._regression)  # ignore 'dF'
        mF, p = F.shape

        if mF != m:
            raise ValueError('Number of rows in F and S do not match.')

        if p > mF:
            raise ArithmeticError('Least-squares problem is undetermined.')

        par = {'corr': self._correlation,
               'regr': self._regression,
               'Y': Y,
               'F': F,
               'D': D,
               'sS': sS}

        self._par = par

    def _objfunc(self, theta: np.ndarray):
        """Maximum likelihood evaluation

        Parameters
        ----------
        theta : np.ndarray
            [description]
        """
        # initialize
        obj = np.Inf

        fitpar = {'sigma2': np.NaN,
                  'beta': np.NaN,
                  'gamma': np.NaN,
                  'C': np.NaN,
                  'Ft': np.NaN,
                  'G': np.NaN, }

        m = self._par['F'].shape[0]

        # set up R
        r = corr(theta, self._par['D'], correlation=self._par['corr'])

        mu = (10 + m) * np.spacing(1)
        R = np.triu(np.ones((m, m)), 1)
        R[R == 1.0] = r
        np.fill_diagonal(R, 1.0 + mu)

        try:
            C = cholesky(R).T
        except LinAlgError:
            # cholesky decomp failed, not positive definite, return inf
            self._fitpar = fitpar
            return obj

        # get least squares solution
        Ft = mldivide(C, self._par['F'])
        Q, G = qr(Ft, mode='economic')

        if 1 / np.linalg.cond(G) < 1e-10:
            # check F
            if np.linalg.cond(self._par['F']) > 1e15:
                raise ValueError(
                    ("F is too ill conditioned. Poor combination of "
                     "regression model and design sites."))
            else:  # matrix Ft is too ill conditioned
                self._fitpar = fitpar
                return obj

        Yt = mldivide(C, self._par['Y'])
        beta = mldivide(G, Q.T @ Yt)
        rho = Yt - Ft @ beta
        sigma2 = np.sum(rho ** 2, axis=0) / m
        detR = np.prod(np.diag(C) ** (2 / m), axis=0)
        obj = np.sum(sigma2, axis=0) * detR

        fitpar['sigma2'] = sigma2 * self._Ysc[1, :] ** 2
        fitpar['beta'] = beta
        fitpar['gamma'] = mrdivide(rho.T, C)
        fitpar['C'] = C
        fitpar['Ft'] = Ft
        fitpar['G'] = G.T

        self._fitpar = fitpar
        return obj

    # ---------------------------- PUBLIC METHODS -----------------------------
    def fit(self, S: np.ndarray, Y: np.ndarray, theta0: np.ndarray,
            lob: np.ndarray = None, upb: np.ndarray = None):
        """ Constrained non-linear least-squares fit of a given correlation 
        model to the provided data set and regression model.

        Parameters
        ----------
        S : (M, N) ndarray
            Design sites: an m-by-n array (m being number of samples and n the
            input dimensions).

        Y : (M, Q) ndarray
            Observed responses: m-by-q array (q is the output dimensions).
            q = 1 means univariate, q > 1 means multivariate.

        theta0 : (N,) ndarray
            Correlation function parameters (`theta`). If `lob` and  `upb` are 
            specified, the value of `theta0` is used as initial guess of the 
            optimization problem. Otherwise, the correlation matrix is 
            calculated with the `theta0` given (no optimization).

        lob : (N,) ndarray, optional
            Lower bound of `theta`.

        upb : (N,) ndarray, optional
            Upper bound of `theta`.

        Raises
        ------
        ValueError
            When `S` and `Y` does not have the same number of rows.

            When specifying only one of either `lob` and `upb`.

            When `lob`, `upb` and `theta0` does not have the same lengths.

            When `theta0` is outside the bound `lob`<`theta`<=`upb`.

            When `theta0` <= 0.

            When `S` has any repeated rows.

        ArithmeticError
            When the optimization fails.

        LinAlgError
            When the design sites are too close to one another resulting in 
            ill-conditioning of the correlation matrix.

        """

        self.S = S
        self.Y = Y

        if theta0.ndim != 1:
            theta0 = theta0.flatten()

        if (lob is None and upb is not None) or \
                (lob is not None and upb is None):
            raise ValueError('You must specify both theta bounds.')

        # load objfunc parameters
        self._initialize()

        if lob is not None and upb is not None:
            # optimization
            if lob.ndim != 1 or upb.ndim != 1:
                lob = lob.flatten()
                upb = upb.flatten()

            lth = theta0.size

            if lob.size != lth or upb.size != lth:
                raise ValueError(
                    'theta0, lob and upb must have the same length.')

            any_lob_le_zero = np.any(np.less_equal(lob, np.zeros(lob.shape)))
            any_upb_lt_lob = np.any(np.less(upb, lob))
            if any_lob_le_zero or any_upb_lt_lob:
                raise ValueError('The bounds must satisfy 0 < lob <= upb.')

            if self._optimizer == 'boxmin':
                # optimize using boxmin
                bmin = BoxMin(self, theta0, lob, upb)

                if np.isinf(bmin.f):
                    raise ValueError(
                        'Bad parameter region. Try increasing upb.')

                f = bmin.f
                theta = bmin.t
                self._objfunc(theta)

                self.theta = theta
                self.perf = bmin.perf_info

            else:
                raise NotImplementedError(
                    "Invalid optimizer option or not implemented.")
        else:  # given theta
            if np.any(np.less_equal(theta0, np.zeros(theta0.shape))):
                raise ValueError('theta0 must be strictly positive.')

            f = self._objfunc(theta0)
            self.theta = theta0

    def predict(self, X: np.ndarray, compute_jacobian=False, compute_mse=False,
                compute_mse_jacobian=False):
        """
        Predictor for y(x) using the given DACE model.

        Parameters
        ----------
        x : (M, N) ndarray
            Trial design sites with n dimensions.

        compute_jacobian : bool, optional
            Whether or not to compute the jacobian.

        compute_mse : bool, optional
            Whether or not to compute the Mean Squared Error (MSE) of the 
            prediction

        compute_mse_jacobian : bool, optional
            Whether or not to compute the MSE jacobian of the prediction.

        Returns
        -------
        y : (M,1) ndarray
            Predicted response at x.

        or1 : (M, ...) ndarray
            If m = 1, and `compute_jacobian` is set to True, then `or1` is a 
            gradient vector/Jacobian matrix of predictor. Otherwise, `or1` is a 
            vector with m rows containing the estimated mean squared error of 
            the predictor.

        or2 : float
            If m = 1, and `compute_mse` is set to True, then `or2` is a the 
            estimated mean squared error (MSE) of the predictor. Otherwise 
            (m > 1 or `compute_mse' set to False), it's a NaN type.

        dmse : (M, ...) ndarray
            The gradient vector/Jacobian Matrix of the MSE. Only available when
            m = 1 and `compute_mse_jacobian` set to True, otherwise it's a NaN 
            type.

        """
        x = X  # to avoid confusion with internal X (train X)

        if x.ndim == 1:  # change from 1d array to 2d
            x = x[np.newaxis, :]

        if x.ndim > 2:
            raise ValueError(
                "Input arrays of dimension higher than 2 aren't allowed.")

        if np.all(np.isnan(self._fitpar['beta'])):
            raise ValueError(
                "Kriging build is invalid because it contains NaN values.")

        m, n = self.m, self.n
        sx = x.shape  # number of trial sites and their dimension

        if np.min(sx) == 1 and n > 1:  # single trial point
            nx = np.max(sx)

            if nx == n:
                mx = 1
                x = x.reshape(1, -1)  # row vector reshaping ´x(:).'´

        else:
            mx, nx = sx

        if nx != n:
            raise ValueError(f"Dimension of trial sites must be {n}.")

        # assign data
        dy, mse, dmse = np.NaN, np.NaN, np.NaN

        # normalize trial sites
        x = (x - self._Ssc[[0], :]) / self._Ssc[[1], :]
        q = self._Ysc.shape[1]  # number of response functions

        if mx == 1:  # one site only
            dx = x - self.S  # distance to design sites

            # get correlation a regression data depending whether or not
            # jacobian info was required
            if compute_jacobian or compute_mse_jacobian:  # jacobian required
                f, df = regrpoly(x, polynomial=self._regression, jacobian=True)
                r, dr = corr(self.theta, dx, correlation=self._correlation,
                             jacobian=True)
            else:
                f = regrpoly(x, polynomial=self._regression)
                r = corr(self.theta, dx, correlation=self._correlation)

            # compute the prediction
            # Scaled predictor
            sy = f @ self._fitpar['beta'] + (self._fitpar['gamma'] @ r).T

            # Predictor
            if q == 1:  # make sure the return is a scalar
                y = np.asscalar((self._Ysc[[0], :] + self._Ysc[[1], :] * sy).T)
            else:  # otherwise, keep it as it is
                y = (self._Ysc[[0], :] + self._Ysc[[1], :] * sy).T

            # compute the prediction jacobian
            if compute_jacobian:
                # scaled jacobian
                sdy = np.transpose(df @ self._fitpar['beta']) + \
                    self._fitpar['gamma'] @ dr

                # unscaled jacobian
                dy = sdy * self._Ysc[[1], :].T / self._Ssc[[1], :]

                if q == 1:  # gradient as column vector for single dimension
                    dy = dy.T

            # compute MSE
            if compute_mse:
                # MSE
                rt = mldivide(self._fitpar['C'], r)
                u = self._fitpar['Ft'].T @ rt - f.T
                v = mldivide(self._fitpar['G'], u)
                mse = np.tile(self._fitpar['sigma2'], (mx, 1)) * \
                    np.tile((1 + np.sum(v ** 2, axis=0) -
                             np.sum(rt ** 2, axis=0)).T,
                            (1, q))

                if q == 1:  # make sure the return is a scalar if q == 1
                    mse = np.asscalar(mse)

                # compute MSE jacobian
                if compute_mse:
                    # scaled gradient as row vector
                    Gv = mldivide(self._fitpar['G'].T, v)
                    g = (self._fitpar['Ft'] @ Gv - rt).T @ \
                        mldivide(self._fitpar['C'], dr) - (df @ Gv).T

                    # unscaled MSE jacobian
                    dmse = np.tile(2 * self._fitpar['sigma2'].reshape(1, -1).T,
                                   (1, nx)) * \
                        np.tile(g / self._Ssc[[1], :], (q, 1))

                    if q == 1:  # gradient as column vector for single dimension
                        dmse = dmse.conj().T

        else:  # several trial sites

            if compute_jacobian or compute_mse_jacobian:  # basic sanitation
                raise ValueError(("Can't compute either prediction or MSE "
                                  "jacobian for several design sites."))

            # Get distance to design sites
            dx = np.zeros((mx * m, n))
            kk = np.arange(m).reshape(1, -1)

            for k in np.arange(mx):
                dx[kk, :] = x[k, :] - self.S
                kk = kk + m

            # Get regression function or correlation
            f = regrpoly(x, polynomial=self._regression)
            r = np.reshape(corr(self.theta, dx, correlation=self._correlation),
                           (m, mx), order='F')

            # scaled predictor
            sy = f @ self._fitpar['beta'] + (self._fitpar['gamma'] @ r).T

            # predictor
            y = self._Ysc[[0], :] + self._Ysc[[1], :] * sy

            # MSE
            if compute_mse:
                rt = mldivide(self._fitpar['C'], r)
                u = mldivide(self._fitpar['G'],
                             self._fitpar['Ft'].T @ rt - f.T)

                mse = self._fitpar['sigma2'] * \
                    (1 + self._colsum(u ** 2) - self._colsum(rt ** 2)).T

        return y, dy, mse, dmse

    def _colsum(self, x):
        """Columnwise sum of elements in x."""

        if x.shape[0] == 1:
            return x

        else:
            return np.sum(x, axis=0)



import numpy as np


def corr(theta, d, correlation='corrgauss', jacobian=False):
    """ Correlation model implementation.

    See page 19 of DACE.pdf explaining on how new equations must be added.

    Parameters
    ----------
    theta : (1, N) or (N, 1) ndarray
        Parameters in the correlation function. The number of elements in `theta` must be equal to the dimension N given
        in `d`. Also, a scalar value is allowed (this corresponds to an isotropic model: all :math:`\\theta_{j}` values
        are the same.
    d : (M,N) ndarray
        M-by-N array containing the differences between design sites.
    correlation: {'corrgauss'}, optional
        Correlation model type. Only Gaussian correlation model implemented.
    jacobian: bool, optional
        Whether or not to compute the jacobian. Be aware that setting this option to True the returned output will be a
        tuple with two elements. Default value is False.

    Returns
    -------
    r : ndarray
        Correlations.
    dr : (M, N) ndarray, optional
        M-by-N array containing the Jacobian information of r. Only calculated when the parameter `jacobian` is set to
        'yes'.

    Raises
    ------
    ValueError
        When the number of elements of `theta` is not n.
        When a invalid `jacobian` option is specified.
    NotImplementedError
        When `correlation` is not a valid option.


    """
    # TODO: implement other correlation models
    if type(jacobian) is not bool:
        raise ValueError(
            "Invalid jacobian option. Valid values are True and False.")

    m, n = d.shape  # number of differences and dimension of data

    if correlation == 'corrgauss':
        if theta.size == 1:
            theta = np.tile(theta, (1, n))
        elif theta.size != n:
            raise ValueError(f'Length of theta must be 1 or {n}.')

        # original -- td = d ** 2 * np.tile(-theta.reshape(-1, 1).T, (m, 1))
        td = d * -theta.flatten()
        # keepdims=True is to force r to be a matrix column
        r = np.exp(np.sum(d * td, axis=1, keepdims=False))

        if jacobian:
            # original -- dr = np.tile(-2 * theta.reshape(-1, 1).T, (m, 1)) * d * np.tile(r, (1, n))
            dr = 2 * td * r[:, np.newaxis]

    else:
        raise NotImplementedError('Correlation not implemented or invalid.')

    if jacobian:
        return r, dr
    else:
        return r



import numpy as np


def regrpoly(S, polynomial='poly0', jacobian=False):
    """Returns the regression matrix (F) and its jacobian (dF) of a given polynomial.

    Parameters
    ----------
    S : numpy.array
        m-by-n array with design sites.
    polynomial : string
        String that defines the polynomial model (default 'poly0' - zero order).
        Valid options are 'poly1' and 'poly2' for first and second orders, respectively.
    jacobian: bool, optional
        Whether or not to compute the jacobian. Be aware that setting this option to True the returned output will be a
        tuple with two elements. Default value is False.

    Returns
    -------
    F: numpy.array
        Regression matrix (F) of the chosen polynomial.
            If `polynomial` == 'poly0', F is a m-by-1 matrix of ones.
            If `polynomial` == 'poly1', F is a m-by-(n+1) matrix.
            If `polynomial` == 'poly2', F is a m-by-(n+1)*(n+1)/2 matrix.
    dF: numpy.array, optional
        Jacobian (dF) of the chosen polynomial. Only calculated when the parameter `jacobian` is set to 'yes'.

    Raises
    ------
    ValueError
        If `polynomial` is an invalid option.
        If `jacobian` is an invalid option.

    """
    if type(jacobian) is not bool:
        raise ValueError(
            "Invalid jacobian option. Valid values are True and False.")

    m, n = S.shape

    # normal info
    if polynomial == 'poly0':
        f = np.ones((m, 1))

    elif polynomial == 'poly1':
        f = np.hstack((np.ones((m, 1)), S))

    elif polynomial == 'poly2':
        nn = int((n + 1) * (n + 2) / 2)
        f = np.hstack((np.ones((m, 1)), S, np.zeros((m, nn - n - 1))))
        j = n + 1
        q = n

        for k in np.arange(1, n + 1):
            # the k:k+1 in S[:,k:k+1] is to extract the column as a Mx1 matrix
            f[:, j + np.arange(q)] = np.tile(S[:, (k - 1):k],
                                             (1, q)) * S[:, np.arange(k - 1, n)]

            j += q
            q -= 1

    else:
        raise ValueError('Invalid regression polynomial choice')

    # jacobian info
    if jacobian:
        if polynomial == 'poly0':
            df = np.zeros((n, 1))

        elif polynomial == 'poly1':
            df = np.hstack((np.zeros((n, 1)), np.eye(n)))

        elif polynomial == 'poly2':
            nn = int((n + 1) * (n + 2) / 2)
            df = np.hstack((np.zeros((n, 1)), np.eye(n),
                            np.zeros((n, nn - n - 1))))
            j = n + 1
            q = n

            for k in np.arange(1, n + 1):
                # the k:k+1 in S[:,k:k+1] is to extract the column as a Mx1 matrix
                df[k - 1, j + np.arange(q)] = np.hstack((2 *
                                                         S[0, (k - 1):k], S[0, np.arange(k, n)]))
                # np.arange(k, n) => k + 1:n

                for i in np.arange(1, n - k + 1):
                    df[k - 1 + i, j + i] = S[0, (k - 1):k]

                j += q
                q -= 1

        else:
            raise ValueError('Invalid regression polynomial choice')

        return f, df

    else:
        return f



import scipy as sp


def mldivide(A, b):
    """Solve systems of linear equations Ax = B for x.

    The matrices A and B must contain the same number of rows.
    If A is a square n-by-n matrix and B is a matrix with n rows,
    then x = A\B is a solution to the equation A*x = B, if it exists.

    Parameters
    ----------
    b : numpy.array
    A : numpy.array

    Returns
    -------
    x : numpy.array
        Solution of Ax = B

    Raises
    ------
    ValueError
        If the number of rows of A and B are not the same

    """

    if A.size == 1 and b.size == 1:
        # perform element wise division, since both of inputs are scalar
        return b / A

    # check if A is square
    m, n = A.shape

    if m != b.shape[0]:
        raise ValueError('The matrices A and b must have the same number of rows.')

    if m == n:  # A is indeed square
        # TODO: implement MATLAB mldivide algorithm
        return sp.linalg.solve(A, b)
    else:
        return sp.linalg.lstsq(A, b)[0]


def mrdivide(B, A):
    """Solve systems of linear equations xA = B for x.

    The matrices A and B must contain the same number of columns.
    If A is a square n-by-n matrix and B is a matrix with n columns,
    then x = B/A is a solution to the equation x*A = B, if it exists.

    Parameters
    ----------
    B : numpy.array
    A : numpy.array

    Returns
    -------
    x : numpy.array
        Solution of xA = B

    Raises
    ------
    ValueError
        If the number of columns of A and B are not the same

    """

    if (A.shape[0] == 1 and A.shape[1] == 1) or (B.shape[0] == 1 and B.shape[1] == 1):
        # perform element wise division, since one or both of inputs are scalar
        return B / A

    if B.shape[1] != A.shape[1]:
        raise ValueError('A and B must contain the same number of columns.')

    return mldivide(A.T, B.T).T



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

PACKAGE_NAME = 'pydace'
VERSION = '0.1.4'

setup(name=PACKAGE_NAME,
      version=VERSION,
      description='Design and Analysis of Computational Experiments as python toolbox.',
      author='Felipe Souza Lima',
      author_email='felipe.lima@eq.ufcg.edu.br',
      url='https://github.com/feslima/pydace',
      license='Apache License 2.0',
      packages=find_packages(where='src', exclude=['tests']),
      package_dir={'': 'src'},
      long_description=long_description,
      long_description_content_type="text/markdown",
      keywords="surrogate, metamodel",
      install_requires=[
          'numpy>=1.16', 
          'scipy>=1.2.0', 
          'pyDOE2>=1.2.1'
          ],
      python_requires='>=3.7',
      classifiers=[
          "Programming Language :: Python :: 3.6",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
          "Topic :: Scientific/Engineering",
      ]
      )
